/******************************************************************************
 *
 * Filename: long.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  4th March 2002
 *
 * Purpose:  Include file for long instructions.
 *
 * (c) Copyright R.H.Thornburrow, 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#ifdef PLATFORM_EPOC32_WINS
	extern int64 _allshl( int64 a, int64 b );
	extern int64 _allshr( int64 a, int64 b );
#endif


void l2d( Frame* fp )
{
	u4 value[2] = { 0, 0 };
	double* d = (double*) value;
	int64* l = (int64*) value;
	int64 l1;

	INSTRUCTION( "%p: l2d\n", fp->pc );
	POP_LONG( fp, value );

	l1 = *l;
	*d = (double) l1;
	PUSH_DOUBLE( fp, value );

	ARITHMETIC_DEBUG( printf( "%p: l2d: 0x%08x%08x => %f\n", fp->pc, value[1], value[0], *d ); )
}


void l2f( Frame* fp )
{
	u4 valuew[2], resultb;
	int64* l = (int64*) valuew;
	float* f = (float*) &resultb;
	float result;

	INSTRUCTION( "%p: l2f\n", fp->pc );
	POP_LONG( fp, valuew );
	result = (float) *l;
	*f = result;
	PUSH_FLOAT( fp, resultb );

	ARITHMETIC_DEBUGN( 2, printf( "%p: l2f: 0x%08x%08x => %f\n", fp->pc, valuew[1], valuew[0], result ); )
}


void l2i( Frame* fp )
{
	u4 value[2];
	int64* l = (int64*) value;
	int32 result;

	INSTRUCTION( "%p: l2i\n", fp->pc );
	POP_LONG( fp, value );
	result = (int32) *l;
	PUSH_INT( fp, result );

	ARITHMETIC_DEBUGN( 2, printf( "%p: l2i: 0x%08x%08x => 0x%08x\n", fp->pc, value[1], value[0], result ); )
}


void ladd( Frame* fp )
{
	u4 value[2];
	int64 *l = (int64*) value;
	int64 value1, value2, result;

	INSTRUCTION( "%p: ladd\n", fp->pc );

	POP_LONG( fp, value );
	value2 = *l;
	POP_LONG( fp, value );
	value1 = *l;
	
	result = value1 + value2;
	
	ZERO_LONG(value);
	
	*l = result;
	PUSH_LONG( fp, value );

	ARITHMETIC_DEBUG( printf( "%p: ladd: %.0f + %.0f => %.0f\n", fp->pc, (double) value1, (double) value2, (double) result ); )
}


int laload( Stack* stp )
{
	Frame *fp = stp->fp;
	int32 index;
	ObjRef* array;
	u4 value[2];

	INSTRUCTION( "%p: laload\n", fp->pc );

	POP_INT( fp, index );
	POP_OBJREF( fp, array );

	if( !array )
		return newCompleteException( stp, NullPointerException, "laload" );

	if( (index < 0) || (index >= (int32) array->length) )
	{
		char buf[32];
		sprintf( buf, "%d", index );
		return newCompleteException( stp, ArrayIndexOutOfBoundsException, buf );
	}

	if( (array->cf != NULL) || (array->type != T_LONG) || (array->data == NULL) )
		return newCompleteException( stp, InternalError, "Bad long[] object reference" );

    ZERO_LONG(value);
    
	value[0] = array->data[index*2];
	value[1] = array->data[index*2+1];

	PUSH_LONG( fp, value );

	ARITHMETIC_DEBUG( printf( "%p: iaload: 0x%p [ %d ] => 0x%08x%08x\n", fp->pc, array->data, index, value[1], value[0] ); )
	return 0;
}


void land( Frame* fp )
{
	u4 value1w[2], value2w[2], resultw[2];

	INSTRUCTION( "%p: land\n", fp->pc );

	POP_LONG( fp, value2w );
	POP_LONG( fp, value1w );
	
	resultw[1] = value1w[1] & value2w[1];
	resultw[0] = value1w[0] & value2w[0];
	
	PUSH_LONG( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: land: 0x%08x%08x & 0x%08x%08x => 0x%08x%08x\n", fp->pc,
			value1w[1], value1w[0], value2w[1], value2w[0], resultw[1], resultw[0] ); )
}


int lastore( Stack* stp )
{
	Frame *fp = stp->fp;
	u4 value[2];
	int32 index;
	ObjRef* array;

	INSTRUCTION( "%p: lastore\n", fp->pc );

	POP_LONG( fp, value );
	POP_INT( fp, index );
	POP_OBJREF( fp, array );

	if( !array )
		return newCompleteException( stp, NullPointerException, "lastore" );

	if( (index < 0) || (index >= (int32) array->length) )
	{
		char buf[32];
		sprintf( buf, "%d", index );
		return newCompleteException( stp, ArrayIndexOutOfBoundsException, buf );
	}

	if( (array->cf != NULL) || (array->type != T_LONG) || (array->data == NULL) )
		return newCompleteException( stp, InternalError, "Bad long[] object reference" );

	array->data[index*2] = value[0];
	array->data[index*2+1] = value[1];

	ARITHMETIC_DEBUG( printf( "%p: lastore: 0x%p [ %d ] <= 0x%08x%08x\n", fp->pc, array->data, index, value[1], value[0] ); )

	return 0;
}


void lcmp( Frame* fp )
{
	u4 value1w[2], value2w[2];
	int64* l1 = (int64*) value1w;
	int64* l2 = (int64*) value2w;
	int64 value1, value2;
	int32 result = 0;

	INSTRUCTION( "%p: lcmp\n", fp->pc );

	POP_LONG( fp, value2w );
	value2 = *l2;

	POP_LONG( fp, value1w );
	value1 = *l1;

	if( value1 > value2 )
		result = 1;
	else if( value1 < value2 )
		result = -1;

	ARITHMETIC_DEBUG( printf( "%p: lcmp: 0x%08x%08x cmp 0x%08x%08x => %d\n", fp->pc,
			value1w[1], value1w[0], value2w[1], value2w[0], result ); )

	PUSH_INT( fp, result );
}


inline void lconst_n( Frame* fp, u1 n )
{
	u4 value[2];
	int64* l = (int64*) value;

	INSTRUCTION( "%p: lconst_%d\n", fp->pc, n );

    ZERO_LONG(value);
    
	*l = (int64) n;
	PUSH_LONG( fp, value );
}


/* note: ldiv is a name clash */
int ldivide( Stack* stp )
{
	u4 valuew1[2], valuew2[2], resultw[2];
	int64* l1 = (int64*) valuew1;
	int64* l2 = (int64*) valuew2;
	int64* l  = (int64*) resultw;
	Frame *fp = stp->fp;
	int64 value1, value2, result;

	INSTRUCTION( "%p: ldiv\n", fp->pc );

	POP_LONG( fp, valuew2 );
	value2 = *l2;
	POP_LONG( fp, valuew1 );
	value1 = *l1;

	if( value2 == 0 )
		return newCompleteException( stp, ArithmeticException, "divide by zero" );

	/* special case - see ldiv documentation */
	if( (value1 == MIN_LONG ) && (value2 == -1) )
	{
		PUSH_LONG( fp, valuew1 );
		return 0;
	}

	result = value1 / value2;
	
	ZERO_LONG(resultw);
	
	*l = result;
	PUSH_LONG( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: ldiv: 0x%08x%08x / 0x%08x%08x => 0x%08x%08x\n", fp->pc,
		valuew1[1], valuew1[0], valuew2[1], valuew2[0], resultw[1], resultw[0] ); )
	return 0;
}


inline void lload_n( Frame* fp, u2 n )
{
	INSTRUCTION( "%p: lload_%d\n", fp->pc, n );
	PUSH_LLONG( fp, n );
}


void lmul( Frame* fp )
{
	u4 value[2];
	int64* l = (int64*) value;
	int64 value1, value2, result;

	INSTRUCTION( "%p: lmul\n", fp->pc );

	POP_LONG( fp, value );
	value2 = *l;

	POP_LONG( fp, value );
	value1 = *l;

    ZERO_LONG(value);

	result = value1 * value2;
	*l = result;

	PUSH_LONG( fp, value );

	ARITHMETIC_DEBUG( printf( "%p: lmul: %.0f * %.0f => %.0f\n", fp->pc, (double) value1, (double) value2, (double) result ); )
}


void lneg( Frame* fp )
{
	u4 valuew[2], resultw[2];
	int64* lv = (int64*) valuew;
	int64* lr = (int64*) resultw;
	int64 value, result;

	INSTRUCTION( "%p: lneg\n", fp->pc );

	POP_LONG( fp, valuew );
	value = *lv;

    ZERO_LONG(resultw);

	result = - value;
	*lr = result;

	PUSH_LONG( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: lneg: 0x%08x%08x => 0x%08x%08x\n", fp->pc,
			valuew[1], valuew[0], resultw[1], resultw[0] ); )
}


void lor( Frame* fp )
{
	u4 value1w[2], value2w[2], resultw[2];

	INSTRUCTION( "%p: lor\n", fp->pc );

	POP_LONG( fp, value2w );
	POP_LONG( fp, value1w );
	
	resultw[1] = value1w[1] | value2w[1];
	resultw[0] = value1w[0] | value2w[0];
	
	PUSH_LONG( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: lor : 0x%08x%08x | 0x%08x%08x => 0x%08x%08x\n", fp->pc,
			value1w[1], value1w[0], value2w[1], value2w[0], resultw[1], resultw[0] ); )
}


int lrem( Stack* stp )
{
	u4 value[2];
	int64* l = (int64*) value;
	Frame *fp = stp->fp;
	int64 value1, value2, result;

	INSTRUCTION( "%p: lrem\n", fp->pc );

	POP_LONG( fp, value );
	value2 = *l;

	POP_LONG( fp, value );
	value1 = *l;

	if( value2 == 0 )
		return newCompleteException( stp, ArithmeticException, "divide by zero" );

    ZERO_LONG(value);

	result = value1 % value2;
	*l = result;

	PUSH_LONG( fp, value );

	ARITHMETIC_DEBUG( printf( "%p: lrem: %.0f %% %.0f => %.0f\n", fp->pc, (double) value1, (double) value2, (double) result ); )
	return 0;
}


int lreturn( Stack* stp )
{
	Frame* fp = stp->fp;
	u4 value[2];
	int stopOnPop = stp->fp->stopOnPop;

	INSTRUCTION( "%p: lreturn\n", fp->pc );

	POP_LONG( fp, value );
	stp->fp = fp->next;
	freeStackFrame( fp, fp->mi );
	fp = stp->fp;

	if( fp )
	{
		fp->prev = NULL;
		PUSH_LONG( fp, value );
	}

#ifdef _DEBUG
	if( fp && method_debug )
	{
		char buf[255];
		sprintMethodInfo( buf, fp->mi );
		printf( "Returned to method %s\n", buf );
	}
#endif /* _DEBUG */

	if( callstack )
		(*callstack)( fp, 0 );

	return ( (!fp || stopOnPop) ? 1 : 0 );
}


void lshl( Frame* fp )
{
	u4 value1w[2], resultw[2];
	int64 *l1 = (int64*) value1w;
	int64 *lr = (int64*) resultw;
	int64 value1, result;
	int32 value2;

	INSTRUCTION( "%p: lshl\n", fp->pc );

	POP_INT( fp, value2 );

	if( value2 > 0x3f )
		printf( "WARNING: lshl being asked to shift by 0x%08x\n", value2 );

	POP_LONG( fp, value1w );
	value1 = *l1;
	
	ZERO_LONG(resultw);
	
#ifdef PLATFORM_EPOC32_WINS
	result = _allshl( value1, (value2 & 0x3f) );
#else
	result = value1 << (value2 & 0x3f);
#endif
	*lr = result;
	PUSH_LONG( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: lshl: 0x%08x%08x << 0x%08x => 0x%08x%08x\n", fp->pc,
			value1w[1], value1w[0], value2, resultw[1], resultw[0] ); )
}


void lshr( Frame* fp )
{
	u4 value1w[2], resultw[2];
	int64 *l1 = (int64*) value1w;
	int64 *lr = (int64*) resultw;
	int64 value1, result;
	int32 value2;

	INSTRUCTION( "%p: lshr\n", fp->pc );

	POP_INT( fp, value2 );

	if( value2 > 0x3f )
		printf( "WARNING: lshr being asked to shift by 0x%08x\n", value2 );

	POP_LONG( fp, value1w );
	value1 = *l1;
	
	ZERO_LONG(resultw);
	
#ifdef PLATFORM_EPOC32_WINS
	result = _allshr( value1, (value2 & 0x3f) );
#else
	result = value1 >> (value2 & 0x3f);
#endif
	*lr = result;
	PUSH_LONG( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: lshr: 0x%08x%08x >> 0x%08x => 0x%08x%08x\n", fp->pc,
			value1w[1], value1w[0], value2, resultw[1], resultw[0] ); )
}


inline void lstore_n( Frame* fp, u2 n )
{
	INSTRUCTION( "%p: lstore_%d\n", fp->pc, n );
	POP_LLONG( fp, n );
}


void lsub( Frame* fp )
{
	u4 value[2];
	int64* l = (int64*) value;
	int64 value1, value2, result;

	INSTRUCTION( "%p: lsub\n", fp->pc );

	POP_LONG( fp, value );
	value2 = *l;

	POP_LONG( fp, value );
	value1 = *l;

    ZERO_LONG(value);

	result = value1 - value2;
	*l = result;

	PUSH_LONG( fp, value );

	ARITHMETIC_DEBUG( printf( "%p: lsub: %.0f - %.0f => %.0f\n",
	    fp->pc, (double) value1, (double) value2, (double) result ); )
}


void lushr( Frame* fp )
{
	u4 value1w[2], resultw[2];
	int64 *l1 = (int64*) value1w;
	int64 *lr = (int64*) resultw;
	int64 value1, result;
	int32 value2;

	INSTRUCTION( "%p: lshr\n", fp->pc );

	POP_INT( fp, value2 );

	if( value2 > 0x3f )
		printf( "WARNING: lshr being asked to shift by 0x%08x\n", value2 );

	POP_LONG( fp, value1w );
	value1 = *l1;

    ZERO_LONG(resultw);

#ifdef PLATFORM_EPOC32_WINS
	if( value1 >= 0 )
		result = _allshr( value1, (value2 & 0x3f) );
	else
		result = _allshr( value1, (value2 & 0x3f) ) + _allshl( 2, ~(value2 & 0x3f) );
#else
	if( value1 >= 0 )
		result = value1 >> (value2 & 0x3f);
	else
		result = (value1 >> (value2 & 0x3f)) + (2 << ~(value2 & 0x3f));
#endif

	*lr = result;
	PUSH_LONG( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: lushr: 0x%08x%08x >> 0x%08x => 0x%08x%08x\n", fp->pc,
			value1w[1], value1w[0], value2, resultw[1], resultw[0] ); )
}


void lxor( Frame* fp )
{
	u4 value1w[2], value2w[2], resultw[2];

	INSTRUCTION( "%p: lxor\n", fp->pc );

	POP_LONG( fp, value2w );
	POP_LONG( fp, value1w );
	
	resultw[1] = value1w[1] ^ value2w[1];
	resultw[0] = value1w[0] ^ value2w[0];
	
	PUSH_LONG( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: lxor: 0x%08x%08x ^ 0x%08x%08x => 0x%08x%08x\n", fp->pc,
			value1w[1], value1w[0], value2w[1], value2w[0], resultw[1], resultw[0] ); )
}


