/******************************************************************************
 *
 * Filename: memory.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Memory management utilities.
 *
 * (c) Copyright R.H.Thornburrow, 1998
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "types.h"
#include "engine.h"
#include "loader.h"
#include "globals.h"
#include "util.h"
#include "hashtable.h"

#include "memory.h"


#ifndef OPTIMIZED
int32 bytes_used = 0;
int32 bytes_peak = 0;
int32 bytes_total = 0;
#endif /* OPTIMIZED */


void* javavm_malloc( size_t size )
{
    void* ret;

#ifndef OPTIMIZED
	bytes_used += size;
	bytes_total += size;

	if( bytes_used > bytes_peak )
		bytes_peak = bytes_used;
#endif /* OPTIMIZED */

	ret = malloc( size );
	
	if( !ret )
	{
	    printf( "FATAL: malloc() failed to allocate %d bytes\n", size );
	    exit( 1 );
	}
	
	return ret;
}


void* javavm_realloc( void* ptr, size_t newsize, size_t oldsize )
{
	void* ret;

#ifndef OPTIMIZED
	bytes_used += newsize - oldsize;

	if( (newsize - oldsize) > 0 )
		bytes_total += newsize - oldsize;

	if( bytes_used > bytes_peak )
		bytes_peak = bytes_used;
#endif /* OPTIMIZED */

	ret = realloc( ptr, newsize );
	
	if( !ret )
	{
	    printf( "FATAL: realloc() failed to allocate %d bytes\n", newsize );
	    exit( 1 );
	}

	return ret;
}


/* DEBUG */
int negative_flag = 0;
/* DEBUG */

void javavm_free( void* ptr, size_t size )
{
#ifndef OPTIMIZED
	bytes_used -= size;

	/* DEBUG */
	if( (bytes_used < 0) && !negative_flag )
	{
		printf( "WARNING: javavm_free() - bytes_used is negative - last size was %d\n", size );
		negative_flag = 1;
	}
	/* DEBUG */
#endif /* OPTIMIZED */

	free( ptr );
}


/**
 * Free up the memory allocated to each of the Java objects we have created.  We
 * walk the linked list "objectPool" and free each object (obj->data) and its
 * associated ObjRef structure (obj).  We also free the linked list element
 * itself, but only after we have walked to the next entry.
 */
void freeObjectPool()
{
	ObjRefList *list = objectPool, *temp;

	while( list )
	{
		javavm_free( list->obj->data, 4 * list->obj->cf->object_size );
		javavm_free( list->obj, sizeof(ObjRef) );
		temp = list;
		list = list->next;
		javavm_free( temp, sizeof(ObjRefList) );
	}
}


/**
 * Free up the memory allocated to each of the Java arrays we have created.  We
 * walk the linked list "arrayPool" and free each array (obj->data) and its
 * associated ObjRef structure (obj).  We also free the linked list element
 * itself, but only after we have walked to the next entry.  The size is
 * calculated only so that javavm_free() can work out how much memory we have
 * used.
 */
void freeArrayPool()
{
	ObjRefList *list = arrayPool, *temp;
	int size;

	while( list )
	{
		size = sizeOfJavaType( list->obj->type );

		if( size ) {
			javavm_free( list->obj->data, size * list->obj->length );
			javavm_free( list->obj, sizeof(ObjRef) );
		} else {
			/* dare not free entry as we don't know what it is! */
			printf( "ERROR: bad array type (%d) - not freeing memory at (%p)\n", list->obj->type, list->obj->data );
		}

		temp = list;
		list = list->next;
		javavm_free( temp, sizeof(ObjRefList) );
	}
}


/**
 * Free up memory allocated to generic blocks of memory that do not form
 * objects, arrays or ClassFile's.  We walk the linked list "memoryBlockPool"
 * and free eack block (the size is recorded as block_size).  We also free the
 * linked list element itself.
 */
void freeMemoryBlockPool()
{
	MemoryBlockList *list = memoryBlockPool, *temp;

	while( list )
	{
		javavm_free( list->block, list->block_size );
		temp = list;
		list = list->next;
		javavm_free( temp, sizeof(MemoryBlockList) );
	}
}


/**
 * Free the memory allocated to an attribute.  The kind field dictates the
 * actual type of the attribute.  Known attributes are handled so that the
 * memory allocated to them and anything they reference (e.g. the byte-code for
 * the Code attribute) is freed.  Unknown attributes are held in memory in a
 * generic block which is the size of an attribute_info record, plus the data
 * for the attribute tacked on the end.  We should never reach the default
 * clause, as this indicates we haven't handled the clean up of a new attribute
 * type.  Note that Code attributes call this function recursively.
 */
void freeAttribute( attribute_info *attr )
{
	int n;

	switch( attr->kind )
	{
	case attr_Unknown:
		javavm_free( attr, sizeof(attribute_info) + attr->attribute_length );
		break;
	case attr_ConstantValue:
		javavm_free( attr, sizeof(ConstantValue_attribute) );
		break;
	case attr_Code:
		{
		Code_attribute *ca = (Code_attribute*) attr;
		/* Exception table */
		for( n = 0; n < ca->exception_table_length; n++ )
			javavm_free( ca->exception_table[n], sizeof(Exception_table) );
		javavm_free( ca->exception_table, ca->exception_table_length * sizeof(Exception_table*) );
		/* Code attributes */
		for( n = 0; n < ca->attributes_count; n++ )
			freeAttribute( ca->attributes[n] );
		javavm_free( ca->attributes, ca->attributes_count * sizeof(attribute_info*) );
		/* The byte-code */
		javavm_free( ca->code, ca->code_length );
		/* This attribute */
		javavm_free( ca, sizeof(Code_attribute) );
		}
		break;
	case attr_SourceFile:
		javavm_free( attr, sizeof(SourceFile_attribute) );
		break;
	case attr_LineNumberTable:
		{
		LineNumberTable_attribute *lnta = (LineNumberTable_attribute*) attr;
		/* Line Number table */
		for( n = 0; n < lnta->line_number_table_length; n++ )
			javavm_free( lnta->line_number_table[n], sizeof(LineNumber_table) );
		javavm_free( lnta->line_number_table, lnta->line_number_table_length * sizeof(LineNumber_table*) );
		/* This attribute */
		javavm_free( lnta, sizeof(LineNumberTable_attribute) );
		}
		break;
	case attr_LocalVariableTable:
		{
		LocalVariableTable_attribute *lvta = (LocalVariableTable_attribute*) attr;
		/* Line Number table */
		for( n = 0; n < lvta->local_variable_table_length; n++ )
			javavm_free( lvta->local_variable_table[n], sizeof(LocalVariable_table) );
		javavm_free( lvta->local_variable_table, lvta->local_variable_table_length * sizeof(LocalVariable_table*) );
		/* This attribute */
		javavm_free( lvta, sizeof(LocalVariableTable_attribute) );
		}
		break;
	default:
		javavm_free( attr, sizeof(attribute_info) /*+ cf->attributes[n]->attribute_length*/ );
		printf( "WARNING: attempt to free class attribute of kind %d\n", attr->kind );
		break;
	}
}


/**
 * Free up the memory allocated to the data associated with each of the
 * ClassFile's we have created (by working through the ClassFile structure).
 * This function acts as a destructor for the pseudo-class ClassFile.  The
 * memory allocated to the ClassFile object itself is free'd by the owning
 * hashtable
 */
void ClassFile_destructor( void* block )
{
	ClassFile *cf = (ClassFile*) block;
	u2 n, m;

	/* Constant pool entries */
	for( n = 1; n < cf->constant_pool_count; n++ )
	{
		switch( cf->constant_pool[n]->tag )
		{
		case CONSTANT_Utf8:
			{
			CONSTANT_Utf8_info *utf8 = (CONSTANT_Utf8_info*) cf->constant_pool[n];
			javavm_free( utf8->utf8_bytes, utf8->utf8_length );
			}
			javavm_free( cf->constant_pool[n], sizeof(CONSTANT_Utf8_info) );
			break;
		case CONSTANT_Integer:
		case CONSTANT_Float:
			javavm_free( cf->constant_pool[n], sizeof(CONSTANT_Integer_info) );
			break;
		case CONSTANT_Long:
		case CONSTANT_Double:
			javavm_free( cf->constant_pool[n], sizeof(CONSTANT_Long_info) );
			n++; /* trailing blank entry */
			break;
		case CONSTANT_Class:       javavm_free( cf->constant_pool[n], sizeof(CONSTANT_Class_info) ); break;
		case CONSTANT_String:      javavm_free( cf->constant_pool[n], sizeof(CONSTANT_String_info) ); break;
		case CONSTANT_Fieldref:    javavm_free( cf->constant_pool[n], sizeof(CONSTANT_Fieldref_info) ); break;
		case CONSTANT_Methodref:   javavm_free( cf->constant_pool[n], sizeof(CONSTANT_Methodref_info) ); break;
		case CONSTANT_InterfaceMethodref: javavm_free( cf->constant_pool[n], sizeof(CONSTANT_InterfaceMethodref_info) ); break;
		case CONSTANT_NameAndType: javavm_free( cf->constant_pool[n], sizeof(CONSTANT_NameAndType_info) ); break;
		default:
			printf( "ERROR: Unknown tag '%d' whilst freeing constant pool\n", cf->constant_pool[n]->tag );
			break;
		}
	}

	javavm_free( cf->constant_pool, cf->constant_pool_count * sizeof(cp_info*) );

	/* Interface entries */
	for( n = 0; n < cf->interfaces_count; n++ )
		javavm_free( cf->interfaces[n], sizeof(InterfaceEntry) );

	javavm_free( cf->interfaces, cf->interfaces_count * sizeof(InterfaceEntry*) );

	/* Fields and field attributes */
	for( n = 0; n < cf->fields_count; n++ )
	{
		/* Field attributes */
		for( m = 0; m < cf->fields[n]->attributes_count; m++ )
			freeAttribute( cf->fields[n]->attributes[m] );

		javavm_free( cf->fields[n]->attributes, cf->fields[n]->attributes_count * sizeof(attribute_info*) );
		javavm_free( cf->fields[n], sizeof(field_info) );
	}

	javavm_free( cf->fields, cf->fields_count * sizeof(field_info*) );

	/* Methods and method attributes */
	for( n = 0; n < cf->methods_count; n++ )
	{
		/* Method attributes */
		for( m = 0; m < cf->methods[n]->attributes_count; m++ )
			freeAttribute( cf->methods[n]->attributes[m] );

		javavm_free( cf->methods[n]->attributes, cf->methods[n]->attributes_count * sizeof(attribute_info*) );
		javavm_free( cf->methods[n], sizeof(method_info) );
	}

	javavm_free( cf->methods, cf->methods_count * sizeof(method_info*) );

	/* Virtual methods - shared with methods if no superclass or superinterface */
	if( cf->super || cf->interfaces_count )
		javavm_free( cf->virtuals, cf->virtuals_count * sizeof(method_info*) );

	/* Class attributes */
	for( n = 0; n < cf->attributes_count; n++ )
		freeAttribute( cf->attributes[n] );

	javavm_free( cf->attributes, cf->attributes_count * sizeof(attribute_info*) );

	/* static data */
	if( cf->static_size && cf->static_data )
		javavm_free( cf->static_data, 4 * cf->static_size );

	/* name of the class (for debugging) */
	javavm_free( cf->name, strlen(cf->name) + 1 );

	/* java.lang.Class object */
	javavm_free( cf->classobj, sizeof(ObjRef) );
}


/**
 * Free up the memory allocated to each of the ClassFile we have created.  The
 * hashtable takes care of freeing the ClassFile itself and the destructor
 * function frees the assosciated data.  The keys are not free'd serparately as
 * they are part of the associated data.
 */
void freeClassFiles()
{
	Hashtable_delete( loadedTable, sizeof(ClassFile), 0, 1, &ClassFile_destructor );
}
