/******************************************************************************
 *
 * Filename: zipfile.h
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  10th May 2003
 *
 * Purpose:  Common definitions for ZIP files.
 *
 * (c) Copyright R.H.Thornburrow, 1998-2003
 *
 * Based on Paul Katz's public domain specification for the PKZIP file format.
 * No proprietry code has been referenced in developing this module.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>

#include "config.h"

#include "zipfile.h"
#include "loader.h"		/* for readU1() and readN() */
#include "memory.h"
 
 
/**
 * Function to create a new ZipCacheEntry.
 */
ZipCacheEntry* newZipCacheEntry( u4 zipofst )
{
	ZipCacheEntry* ret = (ZipCacheEntry*) javavm_malloc( sizeof(ZipCacheEntry) );
	ret->zipofst = zipofst;
	return ret;
}


/* Read a data word into the variable whose address is d */
void readWord( void* d, FILE* f )
{
	u1* data = d;
#ifdef CONFIG_LITTLE_ENDIAN
	*data = fgetc( f ); data++;
	*data = fgetc( f );
#else
	u1 b1 = fgetc( f );
	u1 b2 = fgetc( f );
	*data = b2; data++;
	*data = b1;
#endif
}


/* Read a data double word into the variable whose address is d */
void readDWord( void* d, FILE* f )
{
	u1* data = d;
#ifdef CONFIG_LITTLE_ENDIAN
	*data = fgetc( f ); data++;
	*data = fgetc( f ); data++;
	*data = fgetc( f ); data++;
	*data = fgetc( f );
#else
	u1 b1 = fgetc( f );
	u1 b2 = fgetc( f );
	u1 b3 = fgetc( f );
	u1 b4 = fgetc( f );
	*data = b4; data++;
	*data = b3; data++;
	*data = b2; data++;
	*data = b1;
#endif
}


/* Skip the given number of bytes */
int skipN( FILE* f, u4 count )
{
	return fseek( f, count, SEEK_CUR );
}


/*****************************************************************************/


/**
 * Finds the start of the Central Directory.
 */
int32 findCentralDirectory( FILE *f )
{
	u4 zipsig; /* either ziplocsig or zipcensig */
	ZipEndCentralDir endcds;

	/* (1) Validate that this is a ZIP file with at least one local file. */

	/* Read first double word to check if we have at least one local file */
	readDWord( &zipsig, f );

	if( zipsig != LOCAL_FILE_MAGIC /* local file */ )
	{
		printf( "ZIPFILE: Not a valid ZIP file (Bad Local File magic).\n" );
		return -1;
	}

	/* (2) Read End of Central Directory structure to find the start of the
	 *     Central Directory itself. */

	/* Move to start of End of Central Directory */
	if( fseek( f, -22, SEEK_END ) != 0 )
	{
		printf( "ZIPFILE: Not a valid ZIP file (Can't move to End of Central Dir).\n" );
		return -1;
	}

	/* Read double word to check if we have an End of Central Directory */
	readDWord( &zipsig, f );

	if( zipsig != END_CENTRAL_MAGIC /* end of central dir */ )
	{
		printf( "ZIPFILE: Not a valid ZIP file (Bad end of Central Dir magic).\n" );
		return -1;
	}


	readWord( &endcds.zipedsk, f );	/* Number of this disk */
	readWord( &endcds.zipecen, f );	/* Number of disk with start central dir */

	if( (endcds.zipedsk != endcds.zipecen) || (endcds.zipedsk != 0) )
	{
		printf( "ZIPFILE: Not a valid ZIP file (Disk spanning not allowed).\n" );
		return -1;
	}

	readWord( &endcds.zipenum, f );	/* Total number of entries in central dir on this disk */ 
	readWord( &endcds.zipecenn, f ); /* total number entries in central dir */

	if( endcds.zipenum != endcds.zipecenn )
	{
		printf( "ZIPFILE: Not a valid ZIP file (Central directory spanning not allowed).\n" );
		return -1;
	}

	readDWord( &endcds.zipecsz, f );  /* Size of the central directory */

	readDWord( &endcds.zipeofst, f ); /* Offset of start of central directory with
							   respect to the starting disk number */

	/* (3) Seek to the start of the Central Directory. */

	/* Move to start of Central Directory */
	if( fseek( f, endcds.zipeofst, SEEK_SET ) != 0 )
	{
		printf( "ZIPFILE: Not a valid ZIP file (Can't move to start of Central Dir).\n" );
		return -1;
	}

	return endcds.zipecenn;
}


/**
 * Read a Central Directory record into the given ZipCentralDir structure.
 */
int readCentralDirectory( ZipCentralDir *record, FILE *f )
{
	readU1( &record->zipcver, f );		/* Version made by */
	readU1( &record->zipcos, f );		/* Host operating system */
	readU1( &record->zipcvxt, f );		/* Version needed to extract */
	readU1( &record->zipcexos, f );		/* O/S of version needed for extraction */
	readWord( &record->zipcflg, f );	/* General purpose bit flag */
	readWord( &record->zipcmthd, f );	/* Compression method */
	readWord( &record->zipctim, f );	/* Last mod file time (MS-DOS) */
	readWord( &record->zipcdat, f );	/* Last mod file date (MS-DOS) */
	readDWord( &record->zipccrc, f );	/* CRC-32 */
	readDWord( &record->zipcsiz, f );	/* Compressed size */
	readDWord( &record->zipcunc, f );	/* Uncompressed size */
	readWord( &record->zipcfnl, f );	/* Filename length */
	readWord( &record->zipcxtl, f );	/* Extra field length */
	readWord( &record->zipccml, f );	/* File comment length */
	readWord( &record->zipdsk, f );		/* Disk number start */
	readWord( &record->zipint, f );		/* Internal file attributes */
	readDWord( &record->zipext, f );	/* External file attributes, host system dependent */
	readDWord( &record->zipofst, f );	/* Relative offset of local header */

	/* Check filename buffer is big enough */
	if( record->zipcfnl + 1 >= ZIP_MAX_PATH )
	{
		printf( "ZIPFILE: Unable to read filename - buffer too small (%d).\n", record->zipcfnl );
		return 1;
	}

	readN( record->zipcfn, f, record->zipcfnl ); /* Filename or path */
	record->zipcfn[record->zipcfnl] = '\0';

	if( skipN( f, record->zipcxtl ) != 0 ) /* extra field */
	{
		printf( "ZIPFILE: Unable to read extra field.\n" );
		return 1;
	}

	if( skipN( f, record->zipccml ) != 0 ) /* file comment */
	{
		printf( "ZIPFILE: Unable to read file comment.\n" );
		return 1;
	}

	return 0;
}


/**
 * Read a Local File record into the given ZipLocalFile structure.
 */
int readLocalFile( ZipLocalFile* local, FILE *f )
{
	readWord( &local->zipver, f );
	readWord( &local->zipgenflg, f );
	readWord( &local->zipmthd, f );
	readWord( &local->ziptime, f );
	readWord( &local->zipdate, f );
	readDWord( &local->zipcrc, f );
	readDWord( &local->zipsize, f );
	readDWord( &local->zipuncmp, f );
	readWord( &local->zipfnln, f );
	readWord( &local->zipxtraln, f );

	/* Check filename buffer is big enough */
	if( local->zipfnln + 1 >= ZIP_MAX_PATH )
	{
		printf( "ZIPFILE: Unable to read filename - buffer too small (%d).\n", local->zipfnln );
		return 1;
	}

	readN( local->zipname, f, local->zipfnln );
	local->zipname[local->zipfnln] = '\0';

	if( skipN( f, local->zipxtraln ) != 0 )
	{
		printf( "ZIPFILE: Unable to read extra field.\n" );
		return 1;
	}

	return 0;
}
