/******************************************************************************
 *
 * Filename: java_io_UnixFileSystem.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  8th May 2003
 *
 * Purpose:  Native method implementations for java.io.Win32FileSystem.
 *
 * (c) Copyright R.H.Thornburrow, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <sys/stat.h>    /* for file stat stuff */

#ifdef WIN32
#include <io.h>          /* for file find stuff */
#else
#include <dirent.h>
#endif

#include "../engine/config.h"

#include "java_io_UnixFileSystem.h"

#include "../engine/util.h"
#include "../engine/globals.h"

#ifdef WIN32

#define STAT _stat
#define IS_DIRECTORY(buf)	((buf.st_mode & _S_IFDIR) ? 1 : 0)
#define IS_FILE(buf)		((buf.st_mode & _S_IFREG) ? 1 : 0)
#define MODIFIED_TIME(buf)	buf.st_mtime
#define CAN_READ(buf)		((buf.st_mode & _S_IREAD) ? 1 : 0)
#define SIZE(buf)			buf.st_size

#else

#define STAT stat
#define IS_DIRECTORY(buf)	((buf.st_mode & S_IFDIR) ? 1 : 0)
#define IS_FILE(buf)		((buf.st_mode & S_IFREG) ? 1 : 0)
#define MODIFIED_TIME(buf)	buf.st_mtime
#define CAN_READ(buf)		((buf.st_mode & S_IREAD) ? 1 : 0)
#define SIZE(buf)			buf.st_size

#endif /* WIN32 */


/* JDK 1.4
+   public native java.lang.String canonicalize(java.lang.String)
        throws java.io.IOException;
+   public native int getBooleanAttributes0(java.io.File);
    public native boolean checkAccess(java.io.File, boolean);
    public native long getLastModifiedTime(java.io.File);
+   public native long getLength(java.io.File);
    public native boolean createFileExclusively(java.lang.String)
        throws java.io.IOException;
    public native boolean delete(java.io.File);
    public synchronized native boolean deleteOnExit(java.io.File);
    public native java.lang.String list(java.io.File)[];
    public native boolean createDirectory(java.io.File);
    public native boolean rename(java.io.File, java.io.File);
    public native boolean setLastModifiedTime(java.io.File, long);
    public native boolean setReadOnly(java.io.File);
-   private static native void initIDs();
*/


typedef enum _DotDotState {
	dd_initial = 0, dd_slash, dd_slash_dot, dd_slash_dot_dot, dd_skipping
} DotDotState;


#if JDK_VERSION >= 140
/**
 * Cleans up paths to remove "." and ".." so that path is shortest string
 * possible.
 *
 * We actually only remove "/./" and "/../" as these are most important, but we
 * should really try to remove trailing "/.." and leading "./".  Note that a
 * leading "../" is important.
 *
 * public native java.lang.String canonicalize( java.lang.String path )
 *		throws java.io.IOException;
 */
jobject Java_java_io_UnixFileSystem_canonicalize( JNIEnv* env, jobject self, jobject path )
{
	char input[1024], output[1024];
	int n, len, m = 0;
	DotDotState state = dd_initial;
	jobject ret;

	if( extractFromJavaLangString( input, 1024, path ) )
	{
		/* Work through the input path */
		for( n = 0, len = strlen(input); n < len; n++ )
		{
			/* Work out the next state */
			switch( input[n] )
			{
			case '/':
				switch( state )
				{
				case dd_initial:		/* Just a "/", remeber it */
					state = dd_slash;
					continue;
				case dd_slash_dot:		/* Recognised "/./", reduce to "/" */
					state = dd_slash;
					continue;
				case dd_slash_dot_dot:  /* Recognised "/../" */					
					if( m > 0 )  /* Have we output anything yet? */
					{
						/* Remove previous dir */
						for( ; m >= 0; m-- )
						{
							if( output[m] == '/' )  /* Have we found previous slash */
							{
								state = dd_slash;
								break;  /* Done, go into state dd_slash and go round again */
							}
						}

						if( m < 0 )  /* Start of path */
						{
							/* Go into state dd_initial (so no slash is output) and reset output pos */
							state = dd_initial;
							m = 0;
						}

						continue;  /* Go round  */
					}

					break;  /* No output yet i.e. "^/../", leave alone */
				}
				break;
			case '.':
				switch( state )
				{
				case dd_slash:		/* Now got "/.", remember it */
					state = dd_slash_dot;
					continue;
				case dd_slash_dot:	/* Now got "/..", remember it */
					state = dd_slash_dot_dot;
					continue;
				}
				break;
			}

			/* Whatever we have seen isn't useful, so output any characters we have 'buffered' */
			switch( state )
			{
			case dd_slash:
				output[m++] = '/';
				break;
			case dd_slash_dot:
				output[m++] = '/';
				output[m++] = '.';
				break;
			case dd_slash_dot_dot:
				output[m++] = '/';
				output[m++] = '.';
				output[m++] = '.';
				break;
			}

			/* Deal with the current character, slashes are buffered, all else output */
			switch( input[n] )
			{
			case '/':
				state = dd_slash;
				break;
			default:
				state = dd_initial;
				output[m++] = input[n];
				break;
			}
		}

		/* Output any characters we have 'buffered' */
		switch( state )
		{
		case dd_slash:
			output[m++] = '/';
			break;
		case dd_slash_dot:
			/* Eat trailing "/." */
			break;
		case dd_slash_dot_dot:
			output[m++] = '/';
			output[m++] = '.';
			output[m++] = '.';
			break;
		}

		output[m] = '\0';

		/*printf( "java.io.UnixFileSystem.canonicalize( \"%s\" ) returning \"%s\".\n", input, output );*/

		ret = newJavaLangStringFromBytes( output, m );
		addToList( ret, &objectPool );
		return ret;
	}

	env->retcode = newCompleteException( env, InternalError, "Unable to extract path" );
	return NULL;
}


/* Boolean Attribute bit values */

#define BA_EXISTS    1
#define BA_REGULAR   2
#define BA_DIRECTORY 4
#define BA_HIDDEN    8

/**
 * ???Returns a set of flags indicating the various attributes of the file
 * referred to by 'path.???
 *
 * public native int getBooleanAttributes0( java.io.File file );
 */
jint Java_java_io_UnixFileSystem_getBooleanAttributes0( JNIEnv* env, jobject self, jobject file )
{
	char path[512];
	int32 pathOffset;

	pathOffset = getFieldOffset( file->cf, "path", "Ljava/lang/String;" );

	if( pathOffset < 0 )
	{
		env->retcode = newCompleteException( env, InternalError, "java.io.File does not contain a 'path' field" );
		return 0;
	}

	if( extractFromJavaLangString( path, 512, (ObjRef*) file->data[pathOffset] ) )
	{
		struct STAT buf;
		int okay = STAT( path, &buf );
		jint ret;

		if( okay != 0 )
		{
			/*printf( "java.io.UnixFileSystem.getBooleanAttributes( \"%s\" ) returning 0.\n", path );*/
			return 0;
		}

		ret = BA_EXISTS;

		if( IS_FILE(buf) )		 ret |= BA_REGULAR;
		if( IS_DIRECTORY(buf) )  ret |= BA_DIRECTORY;
		/*if( IS_HIDDEN(buf) )	 ret |= BA_HIDDEN;*/

		/*printf( "java.io.UnixFileSystem.getBooleanAttributes( \"%s\" ) returning %d.\n", path, ret );*/
		return ret;
	}

	env->retcode = newCompleteException( env, InternalError, "Unable to extract path string" );
	return 0;
}


/**
 * ???Get the modification time of 'file' as the number of milliseconds since
 * midnight, January 1, 1970 UTC.???
 *
 * public native long getLastModifiedTime( java.io.File file );
 */
jlong Java_java_io_UnixFileSystem_getLastModifiedTime( JNIEnv* env, jobject self, jobject file )
{
	char path[512];
	int32 pathOffset;

	pathOffset = getFieldOffset( file->cf, "path", "Ljava/lang/String;" );

	if( pathOffset < 0 )
	{
		env->retcode = newCompleteException( env, InternalError, "java.io.File does not contain a 'path' field" );
		return 0;
	}

	if( extractFromJavaLangString( path, 512, (ObjRef*) file->data[pathOffset] ) )
	{
		struct STAT buf;
		int ret = STAT( path, &buf );

		if( ret == 0 )
		{
			/*printf( "java.io.UnixFileSystem.getLastModifiedTime() for '%s' returning to %d.\n", path, MODIFIED_TIME(buf) );*/
			return MODIFIED_TIME(buf);
		}

		/*printf( "java.io.UnixFileSystem.getLastModifiedTime() for '%s' returning to 0.\n", path );*/
		return 0;
	}

	env->retcode = newCompleteException( env, InternalError, "Unable to get path from java.io.File" );
	return 0;
}


/**
 * Returns the extent of the file in bytes.
 *
 * public native long getLength( java.io.File file );
 */
jlong Java_java_io_UnixFileSystem_getLength( JNIEnv* env, jobject self, jobject file )
{
	char path[512];
	int32 pathOffset;

	pathOffset = getFieldOffset( file->cf, "path", "Ljava/lang/String;" );

	if( pathOffset < 0 )
	{
		env->retcode = newCompleteException( env, InternalError, "java.io.File does not contain a 'path' field" );
		return 0;
	}

	if( extractFromJavaLangString( path, 512, (ObjRef*) file->data[pathOffset] ) )
	{
		struct STAT buf;
		int ret = STAT( path, &buf );

		if( ret == 0 )
		{
			/*printf( "java.io.UnixFileSystem.getLength() for '%s' returning to %d.\n", path, SIZE(buf) );*/
			return SIZE(buf);
		}

		/*printf( "java.io.UnixFileSystem.getLength() for '%s' returning to 0.\n", path );*/
		return 0;
	}

	env->retcode = newCompleteException( env, InternalError, "Unable to get path from java.io.File" );
	return 0;
}


/**
 * private static native void initIDs();
 */
void Java_java_io_UnixFileSystem_initIDs( JNIEnv* env, jclass staticself )
{
	/* Do Nothing! */
}




/**
 * ???List all the files in the directory specified by 'dir'.???
 * public native java.lang.String[] list( java.io.File dir );
 */
jobject Java_java_io_UnixFileSystem_list( JNIEnv* env, jobject self, jobject dir )
{
	char pathname[255];
	int pathoff = getFieldOffset( dir->cf, "path", "Ljava/lang/String;" );

	if( pathoff < 0 )
	{
		env->retcode = newCompleteException( env, InternalError, "Unable to get path field from java.io.File" );
		return NULL;
	}

	if( extractFromJavaLangString( pathname, 255, (ObjRef*) dir->data[pathoff] ) )
	{
		ObjRef* array;
		int count = 0;
#ifdef WIN32
		struct _finddata_t c_file;
		long hFile;

			
		strcat( pathname, "\\*.*" );

		/* Find all files in given directory */
		if( (hFile = _findfirst( pathname, &c_file )) != -1L )
		{
			count++;

			/* Find the rest of the files */
			while( _findnext( hFile, &c_file ) == 0 )
				count++;

		   _findclose( hFile );
		}
#else
		DIR *dir;
		struct dirent *entry;

		if( (dir = opendir( pathname )) )
		{
			while( (entry = readdir( dir )) ) {
				count++;
			}
			closedir( dir );
		}
#endif
		array = newObjectRefAndArray( T_OBJREF, count );
		addToList( array, &arrayPool );

		if( count > 0 )
		{
			count = 0;
#ifdef WIN32
			if( (hFile = _findfirst( pathname, &c_file )) != -1L )
			{
				array->data[count] = (u4) newJavaLangStringFromBytes( c_file.name, strlen( c_file.name ) );
				addToList( (ObjRef*) array->data[count], &objectPool );
				count++;

				/* Find the rest of the files */
				while( _findnext( hFile, &c_file ) == 0 )
				{
					array->data[count] = (u4) newJavaLangStringFromBytes( c_file.name, strlen( c_file.name ) );
					addToList( (ObjRef*) array->data[count], &objectPool );
					count++;
				}

			   _findclose( hFile );
			}
#else
			if( (dir = opendir( pathname )) )
			{
				while( (entry = readdir( dir )) )
				{
				    array->data[count] = (u4) newJavaLangStringFromBytes( entry->d_name, strlen( entry->d_name ) );
					addToList( (ObjRef*) array->data[count], &objectPool );
					count++;
				}
				closedir( dir );
			}
#endif
		}

		/*printf( "UnixFileSystem.list( \"%s\" ) returning %d entries.\n", pathname, count );*/

		return array;
	}

	env->retcode = newCompleteException( env, InternalError, "Unable to extract path String" );
	return NULL;
}
#endif /* JDK 1.4 */
