/******************************************************************************
 *
 * Filename: java_lang_Throwable.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  25th March 2002-3
 *
 * Purpose:  Native method implementations for java.lang.Throwable.
 *
 * (c) Copyright R.H.Thornburrow, 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../engine/config.h"
#include "../engine/disasm.h"
#include "../engine/globals.h"
#include "../engine/util.h"
#include "../engine/memory.h"

#include "java_lang_Throwable.h"


/* JDK 1.4 and 1.1.7A
+x   public synchronized native java.lang.Throwable fillInStackTrace();
-1.4 private native int getStackTraceDepth();
-1.4 private native java.lang.StackTraceElement getStackTraceElement(int);
+#   private native void printStackTrace0(java.lang.Object);

	 x = not synchronized in 1.1.7A
	 # = not in JDK 1.4
*/


#if JDK_VERSION >= 140
/**
 * Internal method for creating a new StackTraceElement.
 */
jobject newStackTraceElement( JNIEnv* env, StackTrace_Entry* ste )
{
	ClassFile* stecf = getClass( env, "java/lang/StackTraceElement" );
	jint declaringClassOffset, methodNameOffset, fileNameOffset, lineNumberOffset;
	jobject ret;
	char buf[512];
	int len;

	if( !stecf )
		return NULL;	/* Exception on stack */

	declaringClassOffset = getFieldOffset( stecf, "declaringClass", "Ljava/lang/String;" );
	methodNameOffset     = getFieldOffset( stecf, "methodName", "Ljava/lang/String;" );
	fileNameOffset       = getFieldOffset( stecf, "fileName", "Ljava/lang/String;" );
	lineNumberOffset     = getFieldOffset( stecf, "lineNumber", "I" );

	if( (declaringClassOffset < 0) || (methodNameOffset < 0)
		|| (fileNameOffset < 0) || (lineNumberOffset < 0) )
	{
		env->retcode = newCompleteException( env, InternalError, "StackTraceElement has incorrect fields" );
		return NULL;
	}

	/* Create a new StackTraceElement object */
	ret = newObjectRefAndObject( stecf );
	addToList( ret, &objectPool );

	/* Fill in the class name */
	len = sprintClassName( buf, ste->cf, ste->cf->this_class );
	ret->data[declaringClassOffset] = (u4) newJavaLangStringFromBytes( buf, len );
	addToList( (ObjRef*) ret->data[declaringClassOffset], &objectPool );

	/* Fill in the method name */
	len = sprintSimpleName( buf, ste->cf, ste->mi->name_index );
	ret->data[methodNameOffset]     = (u4) newJavaLangStringFromBytes( buf, len );
	addToList( (ObjRef*) ret->data[methodNameOffset], &objectPool );

	/* Do we have a Source File Attribute?  If so fill in the source file */
	if( ste->mi->cf->sfa )
	{
		len = sprintSimpleName( buf, ste->mi->cf, ste->mi->cf->sfa->sourcefile_index );
		ret->data[fileNameOffset] = (u4) newJavaLangStringFromBytes( buf, len );
		addToList( (ObjRef*) ret->data[fileNameOffset], &objectPool );
	}

	/* Fill in the line number (as calculated when we created the entry) */
	ret->data[lineNumberOffset] = (u4) ste->linenum;

	return ret;
}
#endif /* JDK 1.4 */


/**
 * Internal method for finding the best fit line number for a given pc.
 */
int getBestLineNumber( method_info* mi, int pc )
{
	int best = -1, bestn = -1, n;

	if( !mi->lnta || !mi->lnta->line_number_table_length )
		return -1;

	for( n = 0; n < mi->lnta->line_number_table_length; n++ )
	{
		if( (pc >= mi->lnta->line_number_table[n]->start_pc)
			&& (mi->lnta->line_number_table[n]->start_pc > best) )
		{
			best = mi->lnta->line_number_table[n]->start_pc;
			bestn = n;
		}
	}

	if( (bestn >= 0) && (bestn < mi->lnta->line_number_table_length) )
		return mi->lnta->line_number_table[bestn]->line_number;

	return -1;
}


/**
 * Tags the Throwable (exception) object with a statck trace of the current
 * thread.  This is normally only used internally, but can be used when
 * re-throwing and exception, to make it appear that it came from a different
 * line number or even a different method.  Returns the modified exception
 * object.
 *
 * @since   JDK1.0
 *
 * public native Throwable fillInStackTrace();
 */
extern jobject Java_java_lang_Throwable_fillInStackTrace( JNIEnv* env, jobject self )
#if JDK_VERSION < 140
{
	Frame* baseFrame = env->fp;
	Frame* fp;
	char* buffer = (char*) javavm_malloc( 8096 );
	char *buf = buffer;
	int len;

	/*  Remember block so we can free it */
	addMemoryBlock( buffer, 8096 );

	/* Add each of the frames */
	for( fp = baseFrame; fp; fp = (fp->next ? fp->next : fp->init) )
	{
		len = sprintf( buf, "\tat " ); buf += len;

		if( (fp->next != NULL) && (fp->init != NULL) )
		{
			len = sprintf( buf, "STACK CORRUPT!\n" ); buf += len;
			break;
		}

		if( !fp->mi || !fp->mi->cf )
		{
			len = sprintf( buf, "\n" ); buf += len;
			continue;
		}

		/* Class Name */
		len = sprintClassName( buf, fp->mi->cf, fp->mi->cf->this_class ); buf += len;
		len = sprintf( buf, "." ); buf += len;

		/* Method Name */
		if( CHECKCONSTANT( fp->mi->cf, fp->mi->name_index, CONSTANT_Utf8 ) )
		{
			len = sprintSimpleName( buf, fp->mi->cf, fp->mi->name_index ); buf += len;
		}

		len = sprintf( buf, "(" ); buf += len;

		/* Do we have a Source File Attribute? */
		if( fp->mi->cf->sfa )
		{
			len = sprintSimpleName( buf, fp->mi->cf, fp->mi->cf->sfa->sourcefile_index ); buf += len;
		}

		/* Do we have a Line Number Table? */
		if( fp->mi->lnta )
		{
			int linenum = getBestLineNumber( fp->mi, (fp->pc - fp->min_pc) );

			if( linenum >= 0 )
			{
				len = sprintf( buf, ":%d", linenum ); buf += len;
			}
		}
		
		/* extra position info giving exact pc of exception */
		if( verbose )
		{
			int pc = fp->pc - fp->min_pc;
			len = sprintf( buf, "#%d", pc ); buf += len;
		}

		len = sprintf( buf, ")\n" ); buf += len;
	}

	/* Make sure we didn't go past size of buffer - not a lot we can do now! */
	if( buf > buffer + 8096 )
		printf( "WARNING: Buffer overrun in java.lang.Throwable.fillInStackTrace()\n" );

	/* Fill in stack trace as part of the Throwable object */
	if( self->data )
		self->data[0] = (u4) buffer;

	return self;
}
#else
{
	Frame *fp, *baseFrame = env->fp;
	StackTrace *trace;
	int n = 0, count = 0;

	/* Count the number of stack frames */
	for( fp = baseFrame; fp; fp = (fp->next ? fp->next : fp->init) )
		count++;

	/* Create a new StackTrace */
	trace = (StackTrace*) javavm_malloc( sizeof(StackTrace) );
	trace->entriesc = count;
	trace->entriesv = (StackTrace_Entry*) javavm_malloc( count * sizeof(StackTrace_Entry) );

	/*  Remember blocks so we can free them */
	addMemoryBlock( trace, sizeof(StackTrace) );
	addMemoryBlock( trace->entriesv, count * sizeof(StackTrace_Entry) );

	/* Add each of the frames */
	for( fp = baseFrame; fp; fp = (fp->next ? fp->next : fp->init) )
	{
		trace->entriesv[n].cf      = fp->cf;
		trace->entriesv[n].mi      = fp->mi;
		trace->entriesv[n].pc      = (fp->pc - fp->min_pc);
		trace->entriesv[n].linenum = getBestLineNumber( fp->mi, trace->entriesv[n].pc );
		n++;
	}

	/* Fill in stack trace as part of the Throwable object */
	if( self->data )
		self->data[0] = (u4) trace;

	return self;
}
#endif /* JDK 1.4 */


#if JDK_VERSION >= 140
/**
 * Returns the number of StackFrameElement's representing the stack at the
 * point when this exception was thrown.
 *
 * private native int getStackTraceDepth();
 */
extern jint Java_java_lang_Throwable_getStackTraceDepth( JNIEnv* env, jobject self )
{
	StackTrace *st;

	/* Make sure stack trace has been set */
	if( !self->data || !self->data[0] )
		return 0;

	/* Get the trace */
	st = (StackTrace*) self->data[0];

	/* Return a new StackTraceElement object */
	return st->entriesc;
}


/**
 * Returns the StackTraceElement 'levels' back up the stack at the point when
 * this exception was thrown.
 *
 * private native java.lang.StackTraceElement getStackTraceElement( int levels );
 */
extern jobject Java_java_lang_Throwable_getStackTraceElement( JNIEnv* env, jobject self, jint levels )
{
	StackTrace *st;

	/* Make sure stack trace has been set */
	if( !self->data || !self->data[0] )
		return NULL;

	/* Get the trace */
	st = (StackTrace*) self->data[0];

	/* Check that levels is valid */
	if( (levels < 0) || (levels >= st->entriesc) )
		return NULL;

	/* Return a new StackTraceElement object */
	return newStackTraceElement( env, &st->entriesv[levels] );
}
#endif /* JDK 1.4 */


#if JDK_VERSION < 140
/**
 * Causes any stack trace on the Throwable (this) to be output to the
 * PrintWriter 'stream'.  We don't actually care what type 'stream' is, but the
 * given object must have a void write(String) method.
 *
 * private native void printStackTrace0(Object s);
 */
extern void Java_java_lang_Throwable_printStackTrace0( JNIEnv* env, jobject self, jobject stream )
{
	const char *trace = (const char*) self->data[0];
	env->retcode = 0;

	if( trace )
	{
		ObjRef* str = newJavaLangStringFromBytes( trace, strlen((const char*)trace) );
		addToList( str, &objectPool );
		PUSH_OBJREF( env->fp, stream );
		PUSH_OBJREF( env->fp, str );
		env->retcode = forceExecuteJavaMethod( stream, env, "write", "(Ljava/lang/String;)V" );
	}
	else
	{
		ObjRef* str = newJavaLangStringFromBytes( "null\n", 5 );
		addToList( str, &objectPool );
		PUSH_OBJREF( env->fp, stream );
		PUSH_OBJREF( env->fp, str );
		env->retcode = forceExecuteJavaMethod( stream, env, "write", "(Ljava/lang/String;)V" );
	}
}
#endif /* JDK 1.1 */
