/******************************************************************************
 *
 * Filename: sun_misc_Unsafe.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  3rd May 2003
 *
 * Purpose:  Native method implementations for sun.misc.Unsafe.
 *
 * (c) Copyright R.H.Thornburrow, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>

#include "../engine/config.h"

#include "sun_misc_Unsafe.h"

#include "../engine/memory.h"
#include "../engine/disasm.h"
#include "../engine/engine.h"
#include "../engine/linker.h"
#include "../engine/util.h"


/* JDK 1.4
+   private static native void registerNatives();
+   public native java.lang.Object getObject(java.lang.Object, int);
    public native void putObject(java.lang.Object, int, java.lang.Object);
    public native boolean getBoolean(java.lang.Object, int);
    public native void putBoolean(java.lang.Object, int, boolean);
    public native byte getByte(java.lang.Object, int);
    public native void putByte(java.lang.Object, int, byte);
    public native short getShort(java.lang.Object, int);
    public native void putShort(java.lang.Object, int, short);
    public native char getChar(java.lang.Object, int);
    public native void putChar(java.lang.Object, int, char);
    public native int getInt(java.lang.Object, int);
    public native void putInt(java.lang.Object, int, int);
    public native long getLong(java.lang.Object, int);
    public native void putLong(java.lang.Object, int, long);
    public native float getFloat(java.lang.Object, int);
    public native void putFloat(java.lang.Object, int, float);
    public native double getDouble(java.lang.Object, int);
    public native void putDouble(java.lang.Object, int, double);
+   public native byte getByte(long);
    public native void putByte(long, byte);
    public native short getShort(long);
    public native void putShort(long, short);
    public native char getChar(long);
    public native void putChar(long, char);
    public native int getInt(long);
    public native void putInt(long, int);
    public native long getLong(long);
+   public native void putLong(long, long);
    public native float getFloat(long);
    public native void putFloat(long, float);
    public native double getDouble(long);
    public native void putDouble(long, double);
    public native long getAddress(long);
    public native void putAddress(long, long);
+   public native long allocateMemory(long);
    public native long reallocateMemory(long, long);
    public native void setMemory(long, long, byte);
    public native void copyMemory(long, long, long);
+   public native void freeMemory(long);
+   public native int fieldOffset(java.lang.reflect.Field);
-   public native java.lang.Object staticFieldBase(java.lang.Class);
+   public native void ensureClassInitialized(java.lang.Class);
    public native int arrayBaseOffset(java.lang.Class);
    public native int arrayIndexScale(java.lang.Class);
    public native int addressSize();
    public native int pageSize();
    public native java.lang.Class defineClass(java.lang.String, byte[], int, int, java.lang.ClassLoader, java.security.ProtectionDomain);
    public native java.lang.Class defineClass(java.lang.String, byte[], int, int);
    public native java.lang.Object allocateInstance(java.lang.Class) throws java.lang.InstantiationException;
    public native void monitorEnter(java.lang.Object);
    public native void monitorExit(java.lang.Object);
    public native void throwException(java.lang.Throwable);
*/


#if JDK_VERSION >= 140
/**
 * ???Allocate 'size' bytes of memory and return the address of the block???
 *
 * public native long allocateMemory( long size );
 */
jlong Java_sun_misc_Unsafe_allocateMemory( JNIEnv* env, jobject self, jlong size )
{
	u4 ret = (u4) javavm_malloc( (size_t) size );
	/*printf( "sun.misc.Unsafe.allocateMemory( %d ) returning %p\n", (int32) size, ret );*/
	return (jlong) ret;
}


/**
 * ???Checks to see if a class has been initialized (had its <clinit> method
 * called).  If not, requests that the linker initialize the class.???
 *
 * public native void ensureClassInitialized( java.lang.Class clazz );
 */
void Java_sun_misc_Unsafe_ensureClassInitialized( JNIEnv* env, jobject self, jclass clazz )
{
	ClassFile* cf = (ClassFile*) clazz->data[clazz->cf->object_size];    /* last field is cf */
	
	if( !cf->initialised )
	{
		/*printf( "sun.misc.Unsafe.ensureClassInitialized( %p ) calling initialise()\n", clazz );*/
		initialise( cf, env );
	}
}


/**
 * ???Return the field offset of the given field???
 *
 * public native int fieldOffset( java.lang.reflect.Field field );
 */
jint Java_sun_misc_Unsafe_fieldOffset( JNIEnv* env, jobject self, jobject field )
{
	int32 clazzOffset = getFieldOffset( field->cf, "clazz", "Ljava/lang/Class;" );
	int32 nameOffset  = getFieldOffset( field->cf, "name", "Ljava/lang/String;" );
	ClassFile* cf;
	jclass clazz;
	char name[255];
	u2 n;

	/* Make sure Field contains 'clazz' and 'name' */
	if( (clazzOffset < 0) || (nameOffset < 0) )
	{
		env->retcode = newCompleteException( env, InternalError, "Field does not contain expected fields" );
		return -1;
	}

	/* Extract the parent class and its ClassFile */
	clazz = (jclass) field->data[clazzOffset];
	cf = (ClassFile*) clazz->data[clazz->cf->object_size];    /* last field is cf */

	/* Extract the field name */
	if( extractFromJavaLangString( name, 255, (ObjRef*) field->data[nameOffset] ) )
	{
		/* Look for a matching field name */
		for( n = 0; n < cf->fields_count; n++ )
		{
			if( cf->fields[n] && checkSimpleName( cf, cf->fields[n]->name_index, name ) )
				break;
		}

		/*printf( "sun.misc.Unsafe.fieldOffset( %s.%s ) returning %d\n",
			cf->name, name, (n >= cf->fields_count ? -1 : cf->fields[n]->offset) );*/

		/* Did we find anything? */
		if( n >= cf->fields_count )
			return -1;

		/* Return the field offset */
		return cf->fields[n]->offset;
	}

	env->retcode = newCompleteException( env, InternalError, "Unable to extract field name from String" );
	return -1;
}


/**
 * ???Free the memory at address 'addr'???
 *
 * public native void freeMemory( long addr );
 */
void Java_sun_misc_Unsafe_freeMemory( JNIEnv* env, jobject self, jlong addr )
{
	javavm_free( (void*) (u4) addr, 0 );
}


/**
 * ???Get a byte from the memory at address 'addr'???
 *
 * public native byte getByte( long addr );
 */
jbyte Java_sun_misc_Unsafe_getByte( JNIEnv* env, jobject self, jlong addr )
{
	jbyte* b = (jbyte*) (u4) addr;
	return *b;
}


/**
 * ???Write the long 'value' to the memory at address 'addr'???
 *
 * public native void putLong( long addr, long value );
 */
void Java_sun_misc_Unsafe_putLong( JNIEnv* env, jobject self, jlong addr, jlong value )
{
	jlong* l = (jlong*) (u4) addr;
	*l = value;
}


/**
 * ???Return the Object field of 'obj' at 'offset'.  Can also be used to get
 * static fields and 'obj' will be an object representing the static data???
 *
 * public native java.lang.Object getObject( java.lang.Object obj, int offset );
 */
jobject Java_sun_misc_Unsafe_getObject( JNIEnv* env, jobject self, jobject obj, jint offset )
{
	/* Make sure we have a valid object */
	if( !obj )
	{
		printf( "sun.misc.Unsafe.getObject( %p, %d ) returning null (type is %s).\n", obj, offset, (obj ? (obj->cf ? obj->cf->name : "array") : "null") );
		return NULL;
	}

	/* Check the offset is within bounds (!obj->cf indicates static data) */
	if( (offset < 0)
		|| (obj->cf && (offset >= (jint) obj->cf->object_size))
		|| (!obj->cf && (offset >= (jint) obj->length)) )
	{
		char msg[255];
		sprintf( msg, "Field offset out of bounds (%d)", offset );
		env->retcode = newCompleteException( env, InternalError, msg );
		return NULL;
	}

	/*printf( "sun.misc.Unsafe.getObject( %p (%s), %d ) returning %p (%s).\n",
		obj, (obj ? (obj->cf ? obj->cf->name : "array") : "null"), offset,
		obj->data[offset], (obj->data[offset] ? (((ObjRef*)obj->data[offset])->cf ? ((ObjRef*)obj->data[offset])->cf->name : "array") : "null") );*/
	return (jobject) obj->data[offset];
}


/**
 * private static native void registerNatives();
 */
void Java_sun_misc_Unsafe_registerNatives( JNIEnv* env, jclass staticself )
{
	/* Do Nothing! */
}


/**
 * ???Returns an array (Object) representing the static fields of the given class???
 *
 * public native java.lang.Object staticFieldBase( java.lang.Class clazz );
 */
jobject Java_sun_misc_Unsafe_staticFieldBase( JNIEnv* env, jobject self, jclass clazz )
{
	ClassFile* cf = (ClassFile*) clazz->data[clazz->cf->object_size];    /* last field is cf */
	ClassFile* objcf = getClass( env, "java/lang/Object" );
	ObjRef* ret;

	/* This should never happen, as java.lang.Object is always available */
	if( !objcf )
		return NULL;	/* Exception will be on stack */

	/* Create a new array with cf's static_data as its data
	   and the size of the data as its length */
	ret = (ObjRef*) javavm_malloc( sizeof(ObjRef) );
	ret->cf     = NULL;
	ret->data   = cf->static_data;
	ret->length = cf->static_size;
	ret->type   = 0;

	/*printf( "sun.misc.Unsafe.staticFieldBase( %s ) returning %p.\n", cf->name, ret );*/
	return ret;
}
#endif /* JDK 1.4 */
